/**
 * \addtogroup SD_Card
 * @{
 */

/*****************************************************************************/
/**
 * \file SD_Card.c
 * Sector level SD card routines for the 8051 and Keil C51.
 * \author Copyright (c) 2008, Murray R. Van Luyn. <vanluynm@iinet.net.au>
 * \version 0.0
 * \date 27-11-08
 */
/*****************************************************************************/

/*****************************************************************************
 *
 *     THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 *     OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *     ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 *     DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *     DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 *     GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *     INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *     WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *     NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *     SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************/

#include "SD_Card.h" 


/*****************************************************************************
 *
 *                          Global Variables Definitions  
 *
 *****************************************************************************/
static idata unsigned char cmd[6];

/*****************************************************************************
 *
 *                          Private Function Prototypes   
 *
 *****************************************************************************/
static unsigned char sd_card_write_command(unsigned char * cmd);
static unsigned char sd_card_read_byte(void);
static void sd_card_write_byte(unsigned char byte);
static void sd_card_read_block(unsigned char * cmd, unsigned char * buffer,
                                                    unsigned int bytes);


/*****************************************************************************
 *
 *                        Public Function Implementation  
 *
 *****************************************************************************/

/*****************************************************************************
 *
 *                                sd_card_init()  
 *
 *****************************************************************************/
unsigned char sd_card_init(void)
{
    unsigned int retries = 0u; 
    unsigned char i = 0u;

    /* Configure the port for the SD card */
    SD_CARD_MISO = 1;    // Set as input pin SD_CARD_MISO 
                         // Port outputs are SD_CARD_SCK, SD_CARD_MOSI,
                         // SD_CARD_CS, SD_CARD_SS

    SD_CARD_CS = 1;      // Disable SD card

    for(i = 0u; i < 200u; i++) continue;    // Wait a short time 

#if SD_CARD_HARDWARE_SPI
    SPCON = MSTR | SPEN | SPR0 | SPR2;    // Set SPI to master mode, enabled 
                                          // with divide by 64 clock, SCK 
                                          // low when idle, data sampled on
                                          // the leading edge of SCK.
                                          // (345.6KHz with 22.1184MHz xtal)
                                          // Keep SPI clock rate below 400KHz
                                          // during initialisation.
#endif 

    /* Flush out all SPI data and attempt to get back to an idle bus */
    for(i = 0u; i < 15u; i++)             // Send at least 74 clocks pulses
    {                                     // to the SD card. 
        sd_card_write_byte(0xFFu); 
    } 

    /* Send GO_IDLE_STATE command to reset the SD card */
    cmd[0] = SD_CARD_CMD0;
    cmd[1] = 0x00;
    cmd[2] = 0x00;
    cmd[3] = 0x00;
    cmd[4] = 0x00;
    cmd[5] = SD_CARD_CMD0_CRC; 

    while(SD_CARD_IN_IDLE_STATE != sd_card_write_command(cmd)) 
    { 
        if(retries++ > 10u) 
        { 
            SD_CARD_CS = 1;             // Disable SD card
            return(SD_CARD_ERROR);      // SD card has not answered
        } 
    }

    /* SD card is in SPI mode */

    /* Send CMD55 and ACMD41 pair continuously until card finished startup */
    cmd[5] = 0xFFu;
    retries = 0u;
    i = SD_CARD_NO_RESPONSE;
    do
    {
        /* Send CMD55 to indicate that next command is application specific */
        /* MMC cards do not respond to this command */
        cmd[0] = SD_CARD_CMD55;
        sd_card_write_command(cmd);

        /* Send ACMD41 SEND_OP_COND command */
        /* MMC cards do not respond to this command */
        cmd[0] = SD_CARD_ACMD41;
        i = sd_card_write_command(cmd);

        /* If no response from card then it is not an SD card */
        if(SD_CARD_NO_RESPONSE == i)
        {
            SD_CARD_CS = 1;                  // Disable SD card
            return(SD_CARD_NOT_SD);          // Not an SD card
        }

        /* Timeout if too many tries before card exits idle state */
        if(retries++ > 500u) 
        { 
            SD_CARD_CS = 1;                 // Disable SD card
            return(SD_CARD_ERROR);          // No response
        }

    } while(SD_CARD_IN_IDLE_STATE == i);

#if SD_CARD_HARDWARE_SPI
    /* Set SPI bus to maximum speed */
    SPCON |= SPR0;     // Set SPI to divide by 4 clock
#endif 

    SD_CARD_CS = 1;    // Disable SD card

    return(SD_CARD_NO_ERROR); 
}

/*****************************************************************************
 *
 *                             sd_card_rd_sector()
 *
 *****************************************************************************/
void sd_card_rd_sector(unsigned long sector, unsigned char xdata * buffer)
{
    /* Send READ_SINGLE_BLOCK command to read a block */
    cmd[0] = SD_CARD_CMD17;
    cmd[5] = 0xFFu; 

    /* Integrate block address into read a block command sequence */ 
    sector *= SD_CARD_SECTOR_SIZE;    // sector = sector * SD_CARD_SECTOR_SIZE 

    cmd[1] = (unsigned char)((sector & 0xFF000000L) >> 24u); 
    cmd[2] = (unsigned char)((sector & 0x00FF0000L) >> 16u); 
    cmd[3] = (unsigned char)((sector & 0x0000FF00L) >> 8u); 
    cmd[4] = (unsigned char)(sector & 0x000000FFL);

    sd_card_read_block(cmd, buffer, SD_CARD_SECTOR_SIZE); 

    return; 
}


/*****************************************************************************
 *
 *                             sd_card_wr_sector()
 *
 *****************************************************************************/
void sd_card_wr_sector(unsigned long sector, unsigned char xdata * buffer)
{
    unsigned int i = 0u;

    /* Send WRITE_BLOCK command to write a block to the SD card */ 
    cmd[0] = SD_CARD_CMD24;
    cmd[5] = 0xFFu; 

    /* Integrate block address into write a block command sequence */ 
    sector *= SD_CARD_SECTOR_SIZE;    // sector = sector * SD_CARD_SECTOR_SIZE 

    cmd[1] = (unsigned char)((sector & 0xFF000000L) >> 24u); 
    cmd[2] = (unsigned char)((sector & 0x00FF0000L) >> 16u); 
    cmd[3] = (unsigned char)((sector & 0x0000FF00L) >> 8u); 
    cmd[4] = (unsigned char)(sector & 0x000000FFL);

    /* Issue command and quit early if SD card replies with invalid response */
    if(SD_CARD_NO_ERROR != sd_card_write_command(cmd)) 
    { 
        return; 
    }

    /* Delay a short time whilst clocking the SD card */
    for(i = 0u; i < 100u; i++) 
    { 
        sd_card_read_byte(); 
    } 

    /* Send SD_CARD_START_BLOCK_TOKEN to SD card */
    sd_card_write_byte(SD_CARD_START_BLOCK_TOKEN); 

    /* Write buffer to SD card */
    for(i = 0u; i < SD_CARD_SECTOR_SIZE; i++) 
    { 
        sd_card_write_byte(*buffer++); 
    } 

    /* Send redundant CRC bytes to SD card */
    sd_card_write_byte(0xFFu);
    sd_card_write_byte(0xFFu);

    /* Quit early if SD card replies with write error */
    if((sd_card_read_byte() & 0x1Fu) != 0x05u)    // Response of XXX00101 okay
    {
        return; 
    }

    /* Wait while SD card is busy */
    while(0xFFu != sd_card_read_byte()) continue; 

    /* Disable SD card */
    SD_CARD_CS = 1;

    return; 
}


/*****************************************************************************
 *
 *                             sd_card_read_cid()
 *
 *****************************************************************************/
void sd_card_read_cid(unsigned char xdata * buffer) 
{ 
    /* Send SEND_CID command to read the 16 byte CID register */
    cmd[0] = SD_CARD_CDM10;
    cmd[1] = 0x00u; 
    cmd[2] = 0x00u; 
    cmd[3] = 0x00u; 
    cmd[4] = 0x00u;
    cmd[5] = 0xFFu; 

    sd_card_read_block(cmd, buffer, 16u); 

    return; 
} 


/*****************************************************************************
 *
 *                             sd_card_read_csd()
 *
 *****************************************************************************/
void sd_card_read_csd(unsigned char xdata * buffer) 
{ 
    /* Send SEND_CSD command to read the 16 byte CSD register */
    cmd[0] = SD_CARD_CDM9;
    cmd[1] = 0x00u; 
    cmd[2] = 0x00u; 
    cmd[3] = 0x00u; 
    cmd[4] = 0x00u;
    cmd[5] = 0xFFu; 

    sd_card_read_block(cmd, buffer, 16u); 

    return; 
}


/*****************************************************************************
 *
 *                        Private Function Implementation
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *                           sd_card_write_command()
 *
 *****************************************************************************/
static unsigned char sd_card_write_command(unsigned char * cmd) 
{ 
    unsigned char tmp = 0xFFu; 
    unsigned int timeout = 0u;
    unsigned char i = 0u;

    SD_CARD_CS = 1;                 // Disable SD card

    sd_card_write_byte(0xFFu);      // Send 8 clock pulses 

    SD_CARD_CS = 0;                 // Enable SD card

    /* Send 6 byte command */
    for(i = 0u; i < 6u; i++)    
    { 
        sd_card_write_byte(*cmd++); 
    } 

    /* Wait for a valid response from the SD card */
    while(0xFFu == tmp) 
    { 
        tmp = sd_card_read_byte(); 

        if(timeout++ > 500u) 
        { 
            break;    // No response from the SD card
        } 
    } 

    return(tmp); 
} 


/*****************************************************************************
 *
 *                             sd_card_read_byte() 
 *
 *****************************************************************************/
static unsigned char sd_card_read_byte(void) 
{ 
    unsigned char byte = 0u;

#if SD_CARD_HARDWARE_SPI        // Hardware SPI routine.
    SPDAT = 0xFFu;               // Writing to SPI shift register initiates
                                // transmission.
    while(!SPIF) continue;      // SPIF set when SPI transfer complete.
    byte = SPDAT;               // Read incoming byte from shift register.

#else                           // Software SPI routine.
    idata unsigned char i;
    for(i = 8u; i > 0u; i--)    // Cycle down through bit 
    {                           // positions, MSB to LSB.
        SD_CARD_SCK = 0;        // Generate a clock transition to low. 

        if(SD_CARD_MISO)        // Read the level of the SPI data input pin.
        { 
            byte |= (1 << (i - 1));    // Incoming bit is high so set 
                                       // corresponding bit in 'byte'.
        } 
        else 
        { 
            byte &= ~(1 << (i - 1));   // Incoming bit is low so clear 
                                       // corresponding bit in 'byte'.
        } 

        SD_CARD_SCK = 1;        // Generate a clock transition to high. 
    } 
#endif 

    return(byte); 
} 


/*****************************************************************************
 *
 *                           sd_card_write_byte() 
 *
 *****************************************************************************/
static void sd_card_write_byte(unsigned char byte) 
{ 
#if SD_CARD_HARDWARE_SPI    // Hardware SPI routine.
    SPDAT = byte;           // Writing to SPI shift register initiates
                            // transmission.
    while(!SPIF) continue;  // SPIF set when SPI transfer complete.

#else           
    unsigned char i;          // Software SPI routine.
    for(i = 8u; i > 0u; i--)  // Cycle down through bit 
    {                         // positions, MSB to LSB.

        if(byte & (1 << (i - 1)))    // Is a bit set in 'byte'? 
        { 
            SD_CARD_MOSI = 1;    // Bit set so make SPI data output pin high.
        } 
        else 
        { 
            SD_CARD_MOSI = 0;    // Bit clear so make SPI data output pin low.
        } 

        SD_CARD_SCK = 0;         // Generate a clock transition to low.

        _nop_();

        SD_CARD_SCK = 1;         // Generate a clock transition to high. 
    }

    SD_CARD_MOSI = 1;            // Set SPI data output pin to high again.
#endif 

    return;
} 


/*****************************************************************************
 *
 *                           sd_card_read_block() 
 *
 *****************************************************************************/
static void sd_card_read_block(unsigned char * cmd, unsigned char * buffer,
                                                    unsigned int bytes) 
{ 
    unsigned int i = 0u;

    /* Send command cmd to SD card */
    if (0u != sd_card_write_command(cmd)) 
    { 
        return; 
    } 

    /* Wait for start block token from SD card */
    while(SD_CARD_START_BLOCK_TOKEN != sd_card_read_byte()) continue; 

    /* Read specified number of bytes from the SD card */ 
    for(i = 0u; i < bytes; i++) 
    { 
        *buffer++ = sd_card_read_byte(); 
    } 

    /* Read and disregard CRC bytes */
    sd_card_read_byte();
    sd_card_read_byte();

    /* Disable SD card */
    SD_CARD_CS = 1;      

    return; 
} 


/** @} */


